class ZeusAIException(Exception):
    pass


class PluginError(ZeusAIException):
    pass


class InvalidEndpoint(PluginError):
    pass


class InvalidParams(PluginError):
    pass


class ForbiddenInStandalone(PluginError):
    pass


class InvalidResponse(PluginError):
    pass



"""
When the IO package is run on its own, it runs a basic interactive CLI.
This __main__ module also works as an example of a basic client for other developers
"""
from zeusai_py import io
import getpass


def output(ai_out: str):
    """"""
    print(ai_out)


def main():
    host = input("ZeusAI Server Host: ")
    port = int(input("ZeusAI Server Port: "))
    client = io.Client(host, port)
    username = input("Username: ")
    password = getpass.getpass()
    client.authenticate(username, password)
    client.set_output_func(output)
    client.recv_thread.start()
    while True:
        user_input = input("Input: ")
        client.input(user_input)


if __name__ == "__main__":
    main()



import socket
from zeusai_py.io import exceptions


# Source: https://stackoverflow.com/a/65637828/13800487
class SocketStreamReader:
    def __init__(self, sock: socket.socket):
        self._sock = sock
        self._recv_buffer = bytearray()

    def read(self, num_bytes: int = -1) -> bytes:
        raise NotImplementedError

    def readexactly(self, num_bytes: int) -> bytes:
        buf = bytearray(num_bytes)
        pos = 0
        while pos < num_bytes:
            n = self._recv_into(memoryview(buf)[pos:])
            if n == 0:
                raise exceptions.IncompleteReadError(bytes(buf[:pos]), num_bytes)
            pos += n
        return bytes(buf)

    def readline(self) -> bytes:
        return self.readuntil(b"\n")

    def readuntil(self, separator: bytes = b"\n") -> bytes:
        if len(separator) != 1:
            raise ValueError("Only separators of length 1 are supported.")

        chunk = bytearray(4096)
        start = 0
        buf = bytearray(len(self._recv_buffer))
        bytes_read = self._recv_into(memoryview(buf))
        assert bytes_read == len(buf)

        while True:
            idx = buf.find(separator, start)
            if idx != -1:
                break

            start = len(self._recv_buffer)
            bytes_read = self._recv_into(memoryview(chunk))
            buf += memoryview(chunk)[:bytes_read]

        result = bytes(buf[: idx + 1])
        self._recv_buffer = b"".join(
            (memoryview(buf)[idx + 1 :], self._recv_buffer)
        )
        return result

    def _recv_into(self, view: memoryview) -> int:
        bytes_read = min(len(view), len(self._recv_buffer))
        view[:bytes_read] = self._recv_buffer[:bytes_read]
        self._recv_buffer = self._recv_buffer[bytes_read:]
        if bytes_read == len(view):
            return bytes_read
        bytes_read += self._sock.recv_into(view[bytes_read:])
        return bytes_read



import socket
import json
import threading
from zeusai_py.io import _socket_io


class Client:
    def __init__(self, host: str, port: int) -> None:
        """ The class that handles connecting to and interacting with the ZeusAI server's
        I/O API.

        :param host:
        :param port:
        :return:
        """
        self.host = host
        self.port = port
        self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.conn.connect((self.host, self.port))
        self.reader = _socket_io.SocketStreamReader(self.conn)
        self.output_func = None
        self.recv_thread = threading.Thread(target=self._recv_loop)

    def authenticate(self, username: str, password: str) -> None:
        """ Authenticate with the ZeusAI server.

        :param username:
        :param password:
        :return:
        """
        request_dict = {"endpoint": "auth", "params": {"user": username, "pass": password}}
        self._send_request(request_dict)

    def input(self, input_: str) -> None:
        """ Provide an input to the AI

        :param input_: String containing the user's input.
        :return:
        """
        request_dict = {"endpoint": "input", "params": input_}
        self._send_request(request_dict)

    def set_output_func(self, func) -> None:
        """ Set the output function to func.

        :param func: a function to be called by the recv_thread when the AI provides an output to the user.
            Should take a param for a string containing the AI output.
        :return: None
        """
        self.output_func = func

    def _get_response(self) -> dict:
        response = self.reader.readline()
        response = json.loads(response)
        return response

    def _send_request(self, request_dict: dict) -> None:
        """Serializes request_dict into a JSON bytes object and sends it to the server.

        :param request_dict: Dictionary containing a valid request for the ZeusAI Server API
        :return: None
        """
        serialized_json = json.dumps(request_dict) + "\n"
        serialized_json = serialized_json.encode("utf8")
        self.conn.sendall(serialized_json)

    def _recv_loop(self) -> None:
        """ Loop which runs in a thread to get output from the AI server.
        Takes a set of functions which are called when the associated endpoint is called."""
        while True:
            response = self._get_response()
            if response["endpoint"] == "output":
                self.output_func(response["params"])



import json
import sys

from zeusai_py.io import client
from zeusai_py import exceptions


def _send_request(endpoint: str, params) -> None:
    """Helper function to send a request to the host.

    :param endpoint: String, the endpoint to send the request for.
    :param params: The appropriate params for the endpoint.
    :return: None
    """
    sys.stdout.write(json.dumps({"endpoint": endpoint, "params": params}))
    sys.stdout.flush()


def _get_response() -> dict:
    """Helper function to wait for and returns a response, or raises an exception.

    Note that this blocks until a response is received from the host.

    :raise: InvalidEndpoint, InvalidParams, ForbiddenInStandalone, NotImplementedError
    :return: Dictionary containing the response from the server.
    """
    # Get and load the input
    response = sys.stdin.readline()
    response = json.loads(response)
    if response["endpoint"] == "ERROR":
        params = response["params"]
        error = params["error"]

        # Raise different exceptions depending on the error.
        if error == "invalid endpoint":
            raise exceptions.InvalidEndpoint
        elif error == "invalid params":
            raise exceptions.InvalidParams
        elif error == "forbidden in standalone":
            raise exceptions.ForbiddenInStandalone
        elif error == "not implemented":
            raise NotImplementedError
        else:
            raise exceptions.PluginError
    else:  # If it's not an error:
        # Verify the validity of the response.
        try:
            _ = response["endpoint"]
            _ = response["params"]
        except KeyError:
            raise exceptions.InvalidResponse
        return response


def variable(variables: list) -> dict:
    """Gets a the value of a list of variables from the host.

    :param variables: List of variables to get the values of
    :type variables: List

    :raise: See _get_response().
    :return: Dictionary, containing {key: value} pairs: {variable_name: value}
    """
    _send_request("variable", variables)
    response = _get_response()
    return response["params"]


def followup(question: str) -> None:
    """ Ask the user a followup question, and get their response.

    :param question: The question to ask the user.
    :type question: String

    :raise: See _get_response().
    :return:
    """
    _send_request("followup", question)
    response = _get_response()
    return response["params"]


def simulate_client() -> client.Client:
    """Switches the way requests to and from the server are processed to use the
    I/O API instead of the Plugin API.

    NOTE: THIS IS IRREVERSIBLE! ONCE THE PLUGIN IS USING THE CLIENT API, IT CANNOT
    GO BACK TO THE PLUGIN API.

    :raise: See _get_response().
    :return: Instance of Client
    """
    _send_request("simulate client", "")
    _get_response()
    return client.Client()


def log(level: str, message: str) -> None:
    """Logs `message` with the priority of `level` to the Log File and console, respecting the user's settings.

    `message` is always prefixed with the plugin's name before logging, so it's clear where it's coming from.

    :param level: The logging level, options are: debug
    :type level:
    :param message:
    :type message:

    :raise:
    :return:
    """
    params = {"level": level, "message": message}
    _send_request("log", params)
    _get_response()



from zeusai_py import exceptions


class IOException(exceptions.ZeusAIException):
    """Base exception for all IO module related exceptions"""
    pass


# Source: https://github.com/python/cpython/blob/3.7/Lib/asyncio/streams.py
class IncompleteReadError(EOFError):
    """
    Incomplete read error. Attributes:
    - partial: read bytes string before the end of stream was reached
    - expected: total number of expected bytes (or None if unknown)
    """
    def __init__(self, partial, expected):
        super().__init__(f'{len(partial)} bytes read on a total of '
                         f'{expected!r} expected bytes')
        self.partial = partial
        self.expected = expected

    def __reduce__(self):
        return type(self), (self.partial, self.expected)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeusai_py', 'zeusai_py.io', 'zeusai_py.plugin']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'zeusai-py',
    'version': '0.0.1.0',
    'description': '',
    'long_description': '# ZeusAI.Py\nThe python package to wrap the various APIs associated with ZeusAI.\n\n## Usage for Non-Contributors\nSee the Readthedocs page (TODO) for usage documentation.\n\n## Documentation\nTo view all of the available documentation for this project, please see the [Documents Index](INDEX.md)\nThis includes information on getting started, the development environment, and more.\n\n## Version Numbers\nZeusAI.Py is updated at the same time as the ZeusAI server. The first 3 sections of the ZeusAI.Py version number will be the \nassociated version number for ZeusAI, with the 4th section being the ZeusAI.Py revision number, which is only updated for bugs.\n(i.e. if ZeusAI is on 1.2.3, ZeusAI.Py will be on 1.2.3.x).\n\nIf the first three sections of the ZeusAI.Py version number match your ZeusAI version number, they should be compatable.\n',
    'author': 'JMT',
    'author_email': 'contact@joshinshaker.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/joshinshaker-tech/zeusai_py',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
